import warnings
from pathlib import Path
from unittest.mock import MagicMock, patch

import numpy as np
from PIL import Image as PILImage
from pytest import raises

import menpo.io as mio
from menpo.shape import PointCloud


def test_import_incorrect_built_in():
    with raises(ValueError):
        mio.import_builtin_asset("adskljasdlkajd.obj")


def test_breaking_bad_import():
    img = mio.import_builtin_asset("breakingbad.jpg")
    assert img.shape == (1080, 1920)
    assert img.n_channels == 3
    assert img.landmarks["PTS"].n_points == 68


def test_breaking_bad_import_kwargs():
    img = mio.import_builtin_asset("breakingbad.jpg", normalize=False)
    assert img.pixels.dtype == np.uint8


def test_takeo_import():
    img = mio.import_builtin_asset("takeo.ppm")
    assert img.shape == (225, 150)
    assert img.n_channels == 3
    assert img.landmarks["PTS"].n_points == 68


def test_einstein_import():
    img = mio.import_builtin_asset("einstein.jpg")
    assert img.shape == (1024, 817)
    assert img.n_channels == 1
    assert img.landmarks["PTS"].n_points == 68


def test_lenna_import():
    img = mio.import_builtin_asset("lenna.png")
    assert img.shape == (512, 512)
    assert img.n_channels == 3
    assert img.landmarks.n_groups == 2
    assert img.landmarks["LJSON"].n_points == 68
    assert img.landmarks["pupils"].n_points == 2


def test_import_builtin_ljson():
    lmarks = mio.import_builtin_asset("lenna.ljson")["LJSON"]
    assert lmarks.n_points == 68


def test_import_builtin_pts():
    lmarks = mio.import_builtin_asset("einstein.pts")["PTS"]
    assert lmarks.n_points == 68


def test_resolve_from_paths_single_group():
    def resolver(path):
        test_dict = {"test": path.with_name("takeo.pts")}
        return mio.input.resolve_from_paths(test_dict)

    image = mio.import_image(
        mio.data_path_to("einstein.jpg"), landmark_resolver=resolver
    )
    assert image.landmarks.n_groups == 1
    assert image.landmarks["test"].path == mio.data_path_to("takeo.pts")


def test_resolve_from_paths_multi_group():
    def resolver(path):
        test_dict = {"test": path.with_name("lenna.ljson")}
        return mio.input.resolve_from_paths(test_dict)

    image = mio.import_image(
        mio.data_path_to("einstein.jpg"), landmark_resolver=resolver
    )
    assert image.landmarks.n_groups == 2
    assert set(image.landmarks.keys()) == {"test_LJSON", "test_pupils"}


def test_path():
    # choose a random asset (all should have it!)
    img = mio.import_builtin_asset("einstein.jpg")
    path = mio.data_path_to("einstein.jpg")
    assert img.path == path
    assert img.path.stem == "einstein"
    assert img.path.suffix == ".jpg"
    assert img.path.parent == mio.data_dir_path()
    assert img.path.name == "einstein.jpg"


@patch("menpo.io.input.base._pathlib_glob_for_pattern")
def test_single_suffix_dot_in_path(pathlib_glob):
    import menpo.io.input.base as mio_base
    from pathlib import Path

    fake_path = Path("fake_path.t0.t1.t2")
    pathlib_glob.return_value = [fake_path]
    ext_map = MagicMock()
    ext_map.__contains__.side_effect = lambda x: x == ".t2"

    ret_val = next(mio_base.glob_with_suffix("*.t0.t1.t2", ext_map))
    assert ret_val == fake_path
    ext_map.__contains__.assert_called_with(".t2")


def test_upper_extension_mapped_to_lower():
    import menpo.io.input.base as mio_base
    from pathlib import Path

    ext_map = MagicMock()

    mio_base.importer_for_filepath(Path("fake_path.JPG"), ext_map)
    ext_map.get.assert_called_with(".jpg")


@patch("menpo.io.input.base._pathlib_glob_for_pattern")
def test_double_suffix(pathlib_glob):
    import menpo.io.input.base as mio_base
    from pathlib import Path

    fake_path = Path("fake_path.t1.t2")
    pathlib_glob.return_value = [fake_path]
    ext_map = MagicMock()
    ext_map.__contains__.side_effect = lambda x: x == ".t1.t2"

    ret_val = next(mio_base.glob_with_suffix("*.t1.t2", ext_map))
    assert ret_val == fake_path
    ext_map.__contains__.assert_any_call(".t1.t2")
    ext_map.__contains__.assert_any_call(".t2")


def test_import_image():
    img_path = mio.data_dir_path() / "einstein.jpg"
    im = mio.import_image(img_path)
    assert im.pixels.dtype == np.float
    assert im.n_channels == 1


def test_custom_landmark_resolver():
    def lmark_resolver(path):
        return mio.import_landmark_file(mio.data_path_to("takeo.pts"))

    img = mio.import_image(
        mio.data_path_to("lenna.png"), landmark_resolver=lmark_resolver
    )
    assert img.has_landmarks

    takeo_lmarks = mio.import_builtin_asset.takeo_pts()["PTS"]
    np.allclose(img.landmarks["PTS"].points, takeo_lmarks.points)


def test_landmark_resolver_none():
    img = mio.import_image(mio.data_path_to("lenna.png"), landmark_resolver=None)
    assert not img.has_landmarks


def test_import_image_no_norm():
    img_path = mio.data_dir_path() / "einstein.jpg"
    im = mio.import_image(img_path, normalize=False)
    assert im.pixels.dtype == np.uint8


def test_import_landmark_file():
    lm_path = mio.data_dir_path() / "einstein.pts"
    mio.import_landmark_file(lm_path)


def test_import_images():
    imgs = list(mio.import_images(mio.data_dir_path()))
    imgs_filenames = set(i.path.stem for i in imgs)
    exp_imgs_filenames = {
        "einstein",
        "takeo",
        "tongue",
        "breakingbad",
        "lenna",
        "menpo_thumbnail",
    }
    assert exp_imgs_filenames == imgs_filenames


def test_import_images_are_ordered_and_unduplicated():
    # we know that import_images returns images in path order
    imgs = list(mio.import_images(mio.data_dir_path()))
    imgs_filenames = [i.path.stem for i in imgs]
    print(imgs_filenames)
    exp_imgs_filenames = [
        "breakingbad",
        "einstein",
        "lenna",
        "menpo_thumbnail",
        "takeo",
        "tongue",
    ]
    assert exp_imgs_filenames == imgs_filenames


def test_lsimgs_filenamess():
    assert set(mio.ls_builtin_assets()) == {
        "breakingbad.jpg",
        "einstein.jpg",
        "einstein.pts",
        "lenna.png",
        "breakingbad.pts",
        "lenna.ljson",
        "takeo.ppm",
        "takeo.pts",
        "tongue.jpg",
        "tongue.pts",
        "menpo_thumbnail.jpg",
    }


def test_image_paths():
    ls = mio.image_paths(mio.data_dir_path())
    assert len(list(ls)) == 6


def test_import_images_wrong_path_raises_value_error():
    with raises(ValueError):
        list(mio.import_images("asldfjalkgjlaknglkajlekjaltknlaekstjlakj"))


def test_import_landmark_files_wrong_path_raises_value_error():
    with raises(ValueError):
        list(mio.import_landmark_files("asldfjalkgjlaknglkajlekjaltknlaekstjlakj"))


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_no_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new("RGBA", (10, 15))
    is_file.return_value = True

    im = mio.import_image("fake_image_being_mocked.png", normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 4
    assert im.pixels.dtype == np.uint8


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_RGBA_normalize(is_file, mock_image):
    from menpo.image import MaskedImage

    mock_image.return_value = PILImage.new("RGBA", (10, 15))
    is_file.return_value = True

    im = mio.import_image("fake_image_being_mocked.ppm", normalize=True)
    assert im.shape == (15, 10)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.float
    assert type(im) == MaskedImage


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_RGBA_no_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new("RGBA", (10, 15))
    is_file.return_value = True

    im = mio.import_image("fake_image_being_mocked.ppm", normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 4
    assert im.pixels.dtype == np.uint8


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_L_no_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new("L", (10, 15))
    is_file.return_value = True

    im = mio.import_image("fake_image_being_mocked.ppm", normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.uint8


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_L_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new("L", (10, 15))
    is_file.return_value = True

    im = mio.import_image("fake_image_being_mocked.ppm", normalize=True)
    assert im.shape == (15, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.float


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_I_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new("I", (10, 15))
    is_file.return_value = True
    with raises(ValueError):
        mio.import_image("fake_image_being_mocked.ppm", normalize=True)


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_I_no_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new("I", (10, 15))
    is_file.return_value = True

    im = mio.import_image("fake_image_being_mocked.ppm", normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.int32


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_1_normalize(is_file, mock_image):
    from menpo.image import BooleanImage

    mock_image.return_value = PILImage.new("1", (10, 15))
    is_file.return_value = True

    im = mio.import_image("fake_image_being_mocked.ppm", normalize=True)
    assert im.shape == (15, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.bool
    assert type(im) == BooleanImage


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_1_no_normalize(is_file, mock_image):
    from menpo.image import BooleanImage

    mock_image.return_value = PILImage.new("1", (10, 15))
    is_file.return_value = True

    im = mio.import_image("fake_image_being_mocked.ppm", normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.bool
    assert type(im) == BooleanImage


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_1_proper_conversion(is_file, mock_image):
    from menpo.image import BooleanImage

    arr = np.zeros((10, 10), dtype=np.uint8)
    arr[4, 4] = 255
    mock_image.return_value = PILImage.fromarray(arr).convert("1")
    is_file.return_value = True

    im = mio.import_image("fake_image_being_mocked.ppm", normalize=False)
    assert im.shape == (10, 10)
    assert im.n_channels == 1
    assert im.pixels.dtype == np.bool
    assert type(im) == BooleanImage
    assert np.all(im.pixels == arr.astype(np.bool))


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_P_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new("P", (10, 10))
    is_file.return_value = True

    im = mio.import_image("fake_image_being_mocked.ppm", normalize=True)
    assert im.shape == (10, 10)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.float


@patch("PIL.Image.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_PIL_P_no_normalize(is_file, mock_image):
    mock_image.return_value = PILImage.new("P", (10, 15))
    is_file.return_value = True

    im = mio.import_image("fake_image_being_mocked.ppm", normalize=False)
    assert im.shape == (15, 10)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.uint8


@patch("subprocess.Popen")
@patch("menpo.io.input.video.video_infos_ffprobe")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_ffmpeg_GIF_normalize(is_file, video_infos_ffprobe, pipe):
    video_infos_ffprobe.return_value = {
        "duration": 2,
        "width": 100,
        "height": 150,
        "n_frames": 10,
        "fps": 5,
    }
    empty_frame = np.zeros(150 * 100 * 3, dtype=np.uint8).tobytes()
    pipe.return_value.stdout.read.return_value = empty_frame
    is_file.return_value = True

    ll = mio.import_image("fake_image_being_mocked.gif", normalize=True)
    assert ll.path.name == "fake_image_being_mocked.gif"
    assert ll.fps == 5
    assert len(ll) == 10

    im = ll[0]
    assert im.shape == (150, 100)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.float


@patch("subprocess.Popen")
@patch("menpo.io.input.video.video_infos_ffprobe")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_ffmpeg_GIF_no_normalize(is_file, video_infos_ffprobe, pipe):
    video_infos_ffprobe.return_value = {
        "duration": 2,
        "width": 100,
        "height": 150,
        "n_frames": 10,
        "fps": 5,
    }
    empty_frame = np.zeros(150 * 100 * 3, dtype=np.uint8).tobytes()
    pipe.return_value.stdout.read.return_value = empty_frame
    is_file.return_value = True

    ll = mio.import_image("fake_image_being_mocked.gif", normalize=False)
    assert ll.path.name == "fake_image_being_mocked.gif"
    assert ll.fps == 5
    assert len(ll) == 10

    im = ll[0]
    assert im.shape == (150, 100)
    assert im.n_channels == 3
    assert im.pixels.dtype == np.uint8


@patch("menpo.io.input.landmark.json.load")
@patch("menpo.io.input.base.Path.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_v1_ljson_null_values(is_file, mock_open, mock_dict):
    v1_ljson = {
        "groups": [
            {
                "connectivity": [[0, 1], [1, 2], [2, 3]],
                "label": "chin",
                "landmarks": [
                    {"point": [987.9, 1294.1]},
                    {"point": [96.78, 1246.8]},
                    {"point": [None, 0.1]},
                    {"point": [303.22, 167.2]},
                ],
            },
            {
                "connectivity": [[0, 1]],
                "label": "leye",
                "landmarks": [{"point": [None, None]}, {"point": [None, None]}],
            },
        ],
        "version": 1,
    }
    mock_dict.return_value = v1_ljson
    is_file.return_value = True

    with warnings.catch_warnings(record=True) as found_warnings:
        lmark = mio.import_landmark_file("fake_lmark_being_mocked.ljson", group="LJSON")
    nan_points = np.isnan(lmark.points)

    # Should raise deprecation warning
    assert len(found_warnings) == 1, [w.message for w in found_warnings]
    assert nan_points[2, 0]  # y-coord None point is nan
    assert not nan_points[2, 1]  # x-coord point is not nan
    assert np.all(nan_points[4:, :])  # all of leye label is nan


@patch("menpo.io.input.landmark.json.load")
@patch("menpo.io.input.base.Path.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_v2_ljson_null_values(is_file, mock_open, mock_dict):
    v2_ljson = {
        "labels": [
            {"label": "left_eye", "mask": [0, 1, 2]},
            {"label": "right_eye", "mask": [3, 4, 5]},
        ],
        "landmarks": {
            "connectivity": [[0, 1], [1, 2], [2, 0], [3, 4], [4, 5], [5, 3]],
            "points": [
                [None, 200.5],
                [None, None],
                [316.8, 199.15],
                [339.48, 205.0],
                [358.54, 217.82],
                [375.0, 233.4],
            ],
        },
        "version": 2,
    }

    mock_dict.return_value = v2_ljson
    is_file.return_value = True
    with warnings.catch_warnings(record=True) as w:
        lmark = mio.import_landmark_file("fake_lmark_being_mocked.ljson", group="LJSON")
    nan_points = np.isnan(lmark.points)
    assert nan_points[0, 0]  # y-coord None point is nan
    assert not nan_points[0, 1]  # x-coord point is not nan
    assert np.all(nan_points[1, :])  # all of leye label is nan


@patch("menpo.io.input.landmark.json.load")
@patch("menpo.io.input.base.Path.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_v3_ljson_null_values(is_file, mock_open, mock_dict):
    v3_ljson = {
        "groups": {
            "LJSON": {
                "labels": [
                    {"label": "left_eye", "mask": [0, 1, 2]},
                    {"label": "right_eye", "mask": [3, 4, 5]},
                ],
                "landmarks": {
                    "connectivity": [[0, 1], [1, 2], [2, 0], [3, 4], [4, 5], [5, 3]],
                    "points": [
                        [None, 200.5],
                        [None, None],
                        [316.8, 199.15],
                        [339.48, 205.0],
                        [358.54, 217.82],
                        [375.0, 233.4],
                    ],
                },
            }
        },
        "version": 3,
    }

    mock_dict.return_value = v3_ljson
    is_file.return_value = True
    lmark_dict = mio.import_landmark_file("fake_lmark_being_mocked.ljson")
    assert isinstance(lmark_dict, dict)
    lmark = lmark_dict["LJSON"]
    nan_points = np.isnan(lmark.points)
    assert nan_points[0, 0]  # y-coord None point is nan
    assert not nan_points[0, 1]  # x-coord point is not nan
    assert np.all(nan_points[1, :])  # all of leye label is nan


def test_export_import_pointcloud_ljson_v3(tmpdir):
    tmpdir = Path(tmpdir)
    pc = PointCloud(np.random.random([10, 2]), copy=False)
    mio.export_landmark_file(pc, tmpdir / "test.ljson")
    loaded_pc = mio.import_landmark_file(tmpdir / "test.ljson")["LJSON"]
    np.testing.assert_allclose(pc.points, loaded_pc.points)


@patch("random.shuffle")
def test_shuffle_kwarg_true_calls_shuffle(mock):
    list(mio.import_images(mio.data_dir_path(), shuffle=True))
    assert mock.called


def test_import_as_generator():
    import types

    gen = mio.import_images(mio.data_dir_path(), as_generator=True)
    assert isinstance(gen, types.GeneratorType)
    gen = mio.import_landmark_files(mio.data_dir_path(), as_generator=True)
    assert isinstance(gen, types.GeneratorType)


def test_import_lazy_list():
    from menpo.base import LazyList

    data_path = mio.data_dir_path()
    ll = mio.import_images(data_path)
    assert isinstance(ll, LazyList)
    ll = mio.import_landmark_files(data_path)
    assert isinstance(ll, LazyList)


@patch("menpo.io.input.pickle.pickle.load")
@patch("menpo.io.input.base.Path.open")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_pickle(is_file, mock_open, mock_pickle):
    mock_pickle.return_value = {"test": 1}
    is_file.return_value = True

    objs = mio.import_pickle("mocked.pkl")
    assert isinstance(objs, dict)
    assert "test" in objs
    assert objs["test"] == 1


@patch("menpo.io.input.pickle.pickle.load")
@patch("menpo.io.input.base.Path.open")
@patch("menpo.io.input.base.Path.is_file")
@patch("sys.version_info")
def test_importing_pickle_encoding_py3(version_info, is_file, mock_open, mock_pickle):
    version_info.major = 3
    mock_pickle.return_value = {"test": 1}
    is_file.return_value = True

    mio.import_pickle("mocked.pkl", encoding="latin1")
    assert mock_pickle.call_args[1].get("encoding") == "latin1"


@patch("menpo.io.input.pickle.pickle.load")
@patch("menpo.io.input.base.Path.open")
@patch("menpo.io.input.base.Path.is_file")
@patch("sys.version_info")
def test_importing_pickle_encoding_ignored_py2(
    version_info, is_file, mock_open, mock_pickle
):
    version_info.major = 2
    mock_pickle.return_value = {"test": 1}
    is_file.return_value = True

    mio.import_pickle("mocked.pkl", encoding="latin1")
    assert "encoding" not in mock_pickle.call_args[1]


@patch("menpo.io.input.pickle.pickle.load")
@patch("menpo.io.input.base.Path.open")
@patch("menpo.io.input.base.Path.glob")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_pickles(is_file, glob, mock_open, mock_pickle):
    from pathlib import Path
    from menpo.base import LazyList

    mock_pickle.return_value = {"test": 1}
    is_file.return_value = True
    glob.return_value = [Path("mocked1.pkl"), Path("mocked2.pkl")]

    objs = mio.import_pickles("*")
    assert isinstance(objs, LazyList)
    assert len(objs) == 2
    assert objs[0]["test"] == 1
    assert objs[1]["test"] == 1


@patch("menpo.io.input.pickle.pickle.load")
@patch("menpo.io.input.base.Path.open")
@patch("menpo.io.input.base.Path.glob")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_pickles_as_generator(is_file, glob, mock_open, mock_pickle):
    from pathlib import Path
    import types

    mock_pickle.return_value = {"test": 1}
    is_file.return_value = True
    glob.return_value = [Path("mocked1.pkl"), Path("mocked2.pkl")]

    objs = mio.import_pickles("*", as_generator=True)
    assert isinstance(objs, types.GeneratorType)
    objs = list(objs)
    assert len(objs) == 2
    assert objs[0]["test"] == 1
    assert objs[1]["test"] == 1


@patch("subprocess.Popen")
@patch("menpo.io.input.video.video_infos_ffprobe")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_ffmpeg_avi_no_normalize(is_file, video_infos_ffprobe, pipe):
    video_infos_ffprobe.return_value = {
        "duration": 2,
        "width": 100,
        "height": 150,
        "n_frames": 10,
        "fps": 5,
    }
    is_file.return_value = True
    empty_frame = np.zeros(150 * 100 * 3, dtype=np.uint8).tobytes()
    pipe.return_value.stdout.read.return_value = empty_frame
    ll = mio.import_video("fake_image_being_mocked.avi", normalize=False)
    assert ll.path.name == "fake_image_being_mocked.avi"
    assert ll.fps == 5
    assert len(ll) == 5 * 2
    image = ll[0]
    assert image.shape == ((150, 100))
    assert image.n_channels == 3
    assert image.pixels.dtype == np.uint8


@patch("subprocess.Popen")
@patch("menpo.io.input.video.video_infos_ffprobe")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_ffmpeg_avi_normalize(is_file, video_infos_ffprobe, pipe):
    video_infos_ffprobe.return_value = {
        "duration": 2,
        "width": 100,
        "height": 150,
        "n_frames": 10,
        "fps": 5,
    }
    is_file.return_value = True
    empty_frame = np.zeros(150 * 100 * 3, dtype=np.uint8).tobytes()
    pipe.return_value.stdout.read.return_value = empty_frame
    ll = mio.import_video("fake_image_being_mocked.avi", normalize=True)
    assert ll.path.name == "fake_image_being_mocked.avi"
    assert ll.fps == 5
    assert len(ll) == 5 * 2
    image = ll[0]
    assert image.shape == (150, 100)
    assert image.n_channels == 3
    assert image.pixels.dtype == np.float


@patch("menpo.io.input.video.video_infos_ffprobe")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_ffmpeg_exact_frame_count_no_ffprobe(is_file, video_infos_ffprobe):
    video_infos_ffprobe.side_effect = ValueError
    is_file.return_value = True
    with raises(ValueError):
        mio.import_video(
            "fake_image_being_mocked.avi", normalize=True, exact_frame_count=True
        )


@patch("subprocess.Popen")
@patch("menpo.io.input.video.video_infos_ffprobe")
@patch("menpo.io.input.video.video_infos_ffmpeg")
@patch("menpo.io.input.base.Path.is_file")
def test_importing_ffmpeg_no_exact_frame_count_no_ffprobe(
    is_file, video_infos_ffmpeg, video_infos_ffprobe, pipe
):
    video_infos_ffprobe.side_effect = ValueError
    video_infos_ffmpeg.return_value = {
        "duration": 2,
        "width": 100,
        "height": 150,
        "n_frames": 10,
        "fps": 5,
    }
    is_file.return_value = True
    empty_frame = np.zeros(150 * 100 * 3, dtype=np.uint8).tobytes()
    pipe.return_value.stdout.read.return_value = empty_frame
    ll = mio.import_video(
        "fake_image_being_mocked.avi", normalize=True, exact_frame_count=False
    )
    assert ll.path.name == "fake_image_being_mocked.avi"
    assert ll.fps == 5
    assert len(ll) == 5 * 2
    image = ll[0]
    assert image.shape == (150, 100)
    assert image.n_channels == 3
    assert image.pixels.dtype == np.float


def test_import_images_negative_max_images():
    with raises(ValueError):
        list(mio.import_images(mio.data_dir_path(), max_images=-2))


def test_import_images_zero_max_images():
    with raises(ValueError):
        # different since the conditional 'if max_assets' is skipped,
        # thus all images might be imported.
        list(mio.import_images(mio.data_dir_path(), max_images=0))


# TODO: remove once the normalise argument is removed.
def test_import_image_deprecated_normalise_kwarg():
    with warnings.catch_warnings(record=True) as w:
        img = mio.import_builtin_asset("breakingbad.jpg", normalise=False)
        assert len(w) == 1
    assert img.pixels.dtype == np.uint8


@patch("menpo.io.input.base.Path.is_file")
def test_register_image_importer(is_file):
    from menpo.image import Image

    image = Image.init_blank((10, 10))

    def foo_importer(filepath, **kwargs):
        return image

    is_file.return_value = True

    with patch.dict(mio.input.extensions.image_types, {}, clear=True):
        mio.register_image_importer(".foo", foo_importer)
        new_image = mio.import_image("fake.foo")
    assert image is new_image


@patch("menpo.io.input.base.Path.is_file")
def test_register_landmark_importer(is_file):
    from menpo.shape import PointCloud

    lmark = PointCloud.init_2d_grid((1, 1))

    def foo_importer(filepath, **kwargs):
        return lmark

    is_file.return_value = True

    with patch.dict(mio.input.extensions.image_landmark_types, {}, clear=True):
        mio.register_landmark_importer(".foo", foo_importer)
        new_lmark = mio.import_landmark_file("fake.foo")
    assert lmark is new_lmark


@patch("menpo.io.input.base.Path.is_file")
def test_register_video_importer(is_file):
    from menpo.image import Image
    from menpo.base import LazyList

    def foo_importer(filepath, **kwargs):
        return LazyList([lambda: Image.init_blank((10, 10))])

    is_file.return_value = True

    with patch.dict(mio.input.extensions.ffmpeg_video_types, {}, clear=True):
        mio.register_video_importer(".foo", foo_importer)
        new_video = mio.import_video("fake.foo")
    assert len(new_video) == 1


@patch("menpo.io.input.base.Path.is_file")
def test_register_pickle_importer(is_file):
    obj = object()

    def foo_importer(filepath, **kwargs):
        return obj

    is_file.return_value = True

    with patch.dict(mio.input.extensions.pickle_types, {}, clear=True):
        mio.register_pickle_importer(".foo", foo_importer)
        new_obj = mio.import_pickle("fake.foo")
    assert new_obj is obj


def test_register_no_leading_period():
    ext_map = {}
    mio.input.base._register_importer(ext_map, "foo", lambda x: x)
    assert ".foo" in ext_map
    assert "foo" not in ext_map



from .base import (
    AlignmentViewer2d,
    GraphPlotter,
    ImageViewer,
    ImageViewer2d,
    LandmarkViewer2d,
    LandmarkableViewable,
    Menpo3dMissingError,
    PointGraphViewer2d,
    Renderer,
    Viewable,
    plot_curve,
    plot_gaussian_ellipses,
    view_image_landmarks,
    view_patches,
    viewwrapper,
)
from .textutils import bytes_str, print_dynamic, print_progress, progress_bar_str
from .viewmatplotlib import MatplotlibRenderer



import sys

from setuptools import find_packages, setup


def get_version_and_cmdclass(package_path):
    """Load version.py module without importing the whole package.

    Template code from miniver
    """
    import os
    from importlib.util import module_from_spec, spec_from_file_location

    spec = spec_from_file_location("version", os.path.join(package_path, "_version.py"))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.cmdclass


version, cmdclass = get_version_and_cmdclass("menpo")

# Please see conda/meta.yaml for other binary dependencies
install_requires = ["numpy>=1.14", "scipy>=1.0", "matplotlib>=3.0", "pillow>=4.0"]


setup(
    name="menpo",
    version=version,
    cmdclass=cmdclass,
    description="A Python toolkit for handling annotated data",
    author="The Menpo Team",
    author_email="hello@menpo.org",
    packages=find_packages(),
    install_requires=install_requires,
    package_data={"menpo": ["data/*"]},
    tests_require=["pytest>=6.0", "pytest-mock>=3.0", "black>=20.0"],
)



from .base import imgfeature, ndfeature
from .features import (
    daisy,
    es,
    gaussian_filter,
    gradient,
    igo,
    no_op,
    normalize,
    normalize_norm,
    normalize_std,
    normalize_var,
)

# Optional dependencies may return nothing.
from .optional import *
from .predefined import double_igo
from .visualize import sum_channels



import numpy as np
from numpy.testing import assert_allclose
from pytest import raises

from menpo.feature import gradient
from menpo.image import Image

example_image = np.array([[1.0, 2.0, 6.0], [3.0, 4.0, 5.0]])
y_grad = np.array([[2.0, 2.0, -1.0], [2.0, 2.0, -1.0]])
x_grad = np.array([[1.0, 2.5, 4.0], [1.0, 1.0, 1.0]])


def test_gradient_float():
    dtype = np.float32
    p = example_image.astype(dtype)
    image = Image(p)
    grad_image = gradient(image)
    _check_assertions(grad_image, image.shape, image.n_channels * 2, dtype)
    np_grad = np.gradient(p)
    assert_allclose(grad_image.pixels[0], np_grad[0])
    assert_allclose(grad_image.pixels[1], np_grad[1])


def test_gradient_double():
    dtype = np.float64
    p = example_image.astype(dtype)
    image = Image(p)
    grad_image = gradient(image)
    _check_assertions(grad_image, image.shape, image.n_channels * 2, dtype)
    np_grad = np.gradient(p)
    assert_allclose(grad_image.pixels[0], np_grad[0])
    assert_allclose(grad_image.pixels[1], np_grad[1])


def test_gradient_uint8_exception():
    image = Image(example_image.astype(np.uint8))
    with raises(TypeError):
        gradient(image)


def _check_assertions(actual_image, expected_shape, expected_n_channels, expected_type):
    assert actual_image.pixels.dtype == expected_type
    assert type(actual_image) == Image
    assert actual_image.shape == expected_shape
    assert actual_image.n_channels == expected_n_channels



from collections import OrderedDict
import numpy as np

from .base import labeller_func, validate_input, connectivity_from_array


@labeller_func(group_label="car_streetscene_view_0_8")
def car_streetscene_20_to_car_streetscene_view_0_8(pcloud):
    r"""
    Apply the 8-point semantic labels of "view 0" from the MIT Street Scene
    Car dataset (originally a 20-point markup).

    The semantic labels applied are as follows:

      - front
      - bonnet
      - windshield

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 20
    validate_input(pcloud, n_expected_points)

    front_indices = np.array([0, 1, 3, 2])
    bonnet_indices = np.array([2, 3, 5, 4])
    windshield_indices = np.array([4, 5, 7, 6])

    front_connectivity = connectivity_from_array(front_indices, close_loop=True)
    bonnet_connectivity = connectivity_from_array(bonnet_indices, close_loop=True)
    windshield_connectivity = connectivity_from_array(
        windshield_indices, close_loop=True
    )

    all_connectivity = np.vstack(
        [front_connectivity, bonnet_connectivity, windshield_connectivity]
    )

    mapping = OrderedDict()
    mapping["front"] = front_indices
    mapping["bonnet"] = bonnet_indices
    mapping["windshield"] = windshield_indices

    ind = np.arange(8)
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="car_streetscene_view_1_14")
def car_streetscene_20_to_car_streetscene_view_1_14(pcloud):
    """
    Apply the 14-point semantic labels of "view 1" from the MIT Street Scene
    Car dataset (originally a 20-point markup).

    The semantic labels applied are as follows:

      - front
      - bonnet
      - windshield
      - left_side

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 20
    validate_input(pcloud, n_expected_points)

    front_indices = np.array([0, 1, 3, 2])
    bonnet_indices = np.array([2, 3, 5, 4])
    windshield_indices = np.array([4, 5, 7, 6])
    left_side_indices = np.array([0, 2, 4, 6, 8, 9, 10, 11, 13, 12])

    front_connectivity = connectivity_from_array(front_indices, close_loop=True)
    bonnet_connectivity = connectivity_from_array(bonnet_indices, close_loop=True)
    windshield_connectivity = connectivity_from_array(
        windshield_indices, close_loop=True
    )
    left_side_connectivity = connectivity_from_array(left_side_indices, close_loop=True)

    all_connectivity = np.vstack(
        [
            front_connectivity,
            bonnet_connectivity,
            windshield_connectivity,
            left_side_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["front"] = front_indices
    mapping["bonnet"] = bonnet_indices
    mapping["windshield"] = windshield_indices
    mapping["left_side"] = left_side_indices

    ind = np.hstack((np.arange(9), np.array([10, 12, 14, 16, 18])))
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="car_streetscene_view_2_10")
def car_streetscene_20_to_car_streetscene_view_2_10(pcloud):
    r"""
    Apply the 10-point semantic labels of "view 2" from the MIT Street Scene
    Car dataset (originally a 20-point markup).

    The semantic labels applied are as follows:

      - left_side

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 20
    validate_input(pcloud, n_expected_points)

    left_side_indices = np.array([0, 1, 2, 3, 4, 5, 6, 7, 9, 8])

    left_side_connectivity = connectivity_from_array(left_side_indices, close_loop=True)

    all_connectivity = left_side_connectivity

    mapping = OrderedDict()
    mapping["left_side"] = left_side_indices

    ind = np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18])
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="car_streetscene_view_3_14")
def car_streetscene_20_to_car_streetscene_view_3_14(pcloud):
    r"""
    Apply the 14-point semantic labels of "view 3" from the MIT Street Scene
    Car dataset (originally a 20-point markup).

    The semantic labels applied are as follows:

      - left_side
      - rear windshield
      - trunk
      - rear

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 20
    validate_input(pcloud, n_expected_points)

    left_side_indices = np.array([0, 1, 2, 3, 4, 6, 8, 10, 13, 12])
    rear_windshield_indices = np.array([4, 5, 7, 6])
    trunk_indices = np.array([6, 7, 9, 8])
    rear_indices = np.array([8, 9, 11, 10])

    left_side_connectivity = connectivity_from_array(left_side_indices, close_loop=True)
    rear_windshield_connectivity = connectivity_from_array(
        rear_windshield_indices, close_loop=True
    )
    trunk_connectivity = connectivity_from_array(trunk_indices, close_loop=True)
    rear_connectivity = connectivity_from_array(rear_indices, close_loop=True)

    all_connectivity = np.vstack(
        [
            left_side_connectivity,
            rear_windshield_connectivity,
            trunk_connectivity,
            rear_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["left_side"] = left_side_indices
    mapping["rear_windshield"] = rear_windshield_indices
    mapping["trunk"] = trunk_indices
    mapping["rear"] = rear_indices

    ind = np.array([0, 2, 4, 6, 8, 9, 10, 11, 12, 13, 14, 15, 16, 18])
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="car_streetscene_view_4_14")
def car_streetscene_20_to_car_streetscene_view_4_14(pcloud):
    r"""
    Apply the 14-point semantic labels of "view 4" from the MIT Street Scene
    Car dataset (originally a 20-point markup).

    The semantic labels applied are as follows:

      - front
      - bonnet
      - windshield
      - right_side

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 20
    validate_input(pcloud, n_expected_points)

    front_indices = np.array([0, 1, 3, 2])
    bonnet_indices = np.array([2, 3, 5, 4])
    windshield_indices = np.array([4, 5, 7, 6])
    right_side_indices = np.array([8, 9, 10, 11, 13, 12, 1, 3, 5, 7])

    front_connectivity = connectivity_from_array(front_indices, close_loop=True)
    bonnet_connectivity = connectivity_from_array(bonnet_indices, close_loop=True)
    windshield_connectivity = connectivity_from_array(
        windshield_indices, close_loop=True
    )
    right_side_connectivity = connectivity_from_array(
        right_side_indices, close_loop=True
    )

    all_connectivity = np.vstack(
        [
            front_connectivity,
            bonnet_connectivity,
            windshield_connectivity,
            right_side_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["front"] = front_indices
    mapping["bonnet"] = bonnet_indices
    mapping["windshield"] = windshield_indices
    mapping["right_side"] = right_side_indices

    ind = np.hstack([np.arange(8), np.array([9, 11, 13, 15, 17, 19])])
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="car_streetscene_view_5_10")
def car_streetscene_20_to_car_streetscene_view_5_10(pcloud):
    r"""
    Apply the 10-point semantic labels of "view 5" from the MIT Street Scene
    Car dataset (originally a 20-point markup).

    The semantic labels applied are as follows:

      - right_side

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 20
    validate_input(pcloud, n_expected_points)

    right_side_indices = np.array([0, 1, 2, 3, 4, 5, 6, 7, 9, 8])

    right_side_connectivity = connectivity_from_array(
        right_side_indices, close_loop=True
    )

    all_connectivity = right_side_connectivity

    mapping = OrderedDict()
    mapping["right_side"] = right_side_indices

    ind = np.array([1, 3, 5, 7, 9, 11, 13, 15, 17, 19])
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="car_streetscene_view_6_14")
def car_streetscene_20_to_car_streetscene_view_6_14(pcloud):
    r"""
    Apply the 14-point semantic labels of "view 6" from the MIT Street Scene
    Car dataset (originally a 20-point markup).

    The semantic labels applied are as follows:

      - right_side
      - rear_windshield
      - trunk
      - rear

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 20
    validate_input(pcloud, n_expected_points)

    right_side_indices = np.array([0, 1, 2, 3, 5, 7, 9, 11, 13, 12])
    rear_windshield_indices = np.array([4, 5, 7, 6])
    trunk_indices = np.array([6, 7, 9, 8])
    rear_indices = np.array([8, 9, 11, 10])

    right_side_connectivity = connectivity_from_array(
        right_side_indices, close_loop=True
    )
    rear_windshield_connectivity = connectivity_from_array(
        rear_windshield_indices, close_loop=True
    )
    trunk_connectivity = connectivity_from_array(trunk_indices, close_loop=True)
    rear_connectivity = connectivity_from_array(rear_indices, close_loop=True)

    all_connectivity = np.vstack(
        [
            right_side_connectivity,
            rear_windshield_connectivity,
            trunk_connectivity,
            rear_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["right_side"] = right_side_indices
    mapping["rear_windshield"] = rear_windshield_indices
    mapping["trunk"] = trunk_indices
    mapping["rear"] = rear_indices

    ind = np.array([1, 3, 5, 7, 8, 9, 10, 11, 12, 13, 14, 15, 17, 19])
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="car_streetscene_view_7_8")
def car_streetscene_20_to_car_streetscene_view_7_8(pcloud):
    r"""
    Apply the 8-point semantic labels of "view 7" from the MIT Street Scene
    Car dataset (originally a 20-point markup).

    The semantic labels applied are as follows:

      - rear_windshield
      - trunk
      - rear

    References
    ----------
    .. [1] http://www.cs.cmu.edu/~vboddeti/alignment.html
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 20
    validate_input(pcloud, n_expected_points)

    rear_windshield_indices = np.array([0, 1, 3, 2])
    trunk_indices = np.array([2, 3, 5, 4])
    rear_indices = np.array([4, 5, 7, 6])

    rear_windshield_connectivity = connectivity_from_array(
        rear_windshield_indices, close_loop=True
    )
    trunk_connectivity = connectivity_from_array(trunk_indices, close_loop=True)
    rear_connectivity = connectivity_from_array(rear_indices, close_loop=True)

    all_connectivity = np.vstack(
        [rear_windshield_connectivity, trunk_connectivity, rear_connectivity]
    )

    mapping = OrderedDict()
    mapping["rear_windshield"] = rear_windshield_indices
    mapping["trunk"] = trunk_indices
    mapping["rear"] = rear_indices

    ind = np.arange(8, 16)
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping
    )

    return new_pcloud, mapping



from os.path import sep as PATH_SEP
from pathlib import Path
from unittest.mock import patch

from pytest import raises

from menpo.io.input.base import _pathlib_glob_for_pattern
from menpo.io.output.base import _parse_and_validate_extension


@patch("menpo.io.input.base.Path.glob")
def test_glob_parse_contains_file_glob_no_sort(mock_glob):
    path = "/tmp/test.*"
    mock_glob.return_value = ["/tmp/test.test"]
    result = list(_pathlib_glob_for_pattern(path, sort=False))
    assert len(result) == 1
    mock_glob.assert_called_with("test.*")


@patch("menpo.io.input.base.Path.glob")
def test_glob_parse_contains_dir_glob_no_sort(mock_glob):
    path = "/tmp/**/*"
    mock_glob.return_value = ["/tmp/a/test.test", "/tmp/b/test.test"]
    result = list(_pathlib_glob_for_pattern(path, sort=False))
    assert len(result) == 2
    mock_glob.assert_called_with("**/*".replace("/", PATH_SEP))


@patch("menpo.io.input.base.Path.glob")
def test_glob_parse_sort(mock_glob):
    path = "/tmp/**/*"
    mock_glob.return_value = ["/tmp/b/test.test", "/tmp/a/test.test"]
    result = list(_pathlib_glob_for_pattern(path, sort=True))
    assert len(result) == 2
    assert result[0] == mock_glob.return_value[1]
    mock_glob.assert_called_with("**/*".replace("/", PATH_SEP))


def test_parse_extension_given_extension():
    filepath = "test.jpg"
    extension = filepath[-4:]
    parsed_extension = _parse_and_validate_extension(
        Path(filepath), extension, {".jpg": None}
    )
    assert parsed_extension == extension


def test_parse_extension_no_given_extension():
    filepath = "test.jpg"
    extension = filepath[-4:]
    parsed_extension = _parse_and_validate_extension(
        Path(filepath), None, {".jpg": None}
    )
    assert parsed_extension == extension


def test_parse_double_extension_given_extension():
    filepath = "test.pkl.gz"
    extension = filepath[-7:]
    parsed_extension = _parse_and_validate_extension(
        Path(filepath), extension, {".pkl.gz": None}
    )
    assert parsed_extension == extension


def test_parse_double_extension_no_extension():
    filepath = "test.pkl.gz"
    extension = filepath[-7:]
    parsed_extension = _parse_and_validate_extension(
        Path(filepath), None, {".pkl.gz": None}
    )
    assert parsed_extension == extension


def test_parse_period_in_name_given_extension():
    filepath = "test.1.jpg"
    extension = filepath[-4:]
    parsed_extension = _parse_and_validate_extension(
        Path(filepath), extension, {".jpg": None}
    )
    assert parsed_extension == extension


def test_parse_period_in_name_no_extension():
    filepath = "test.1.jpg"
    extension = filepath[-4:]
    parsed_extension = _parse_and_validate_extension(
        Path(filepath), None, {".jpg": None}
    )
    assert parsed_extension == extension


def test_parse_period_in_name_double_given_extension():
    filepath = "test.1.pkl.gz"
    extension = filepath[-7:]
    parsed_extension = _parse_and_validate_extension(
        Path(filepath), extension, {".pkl.gz": None}
    )
    assert parsed_extension == extension


def test_parse_period_in_name_double_no_extension():
    filepath = "test.1.pkl.gz"
    extension = filepath[-7:]
    parsed_extension = _parse_and_validate_extension(
        Path(filepath), extension, {".pkl.gz": None}
    )
    assert parsed_extension == extension


def test_parse_unknown_extension_raises_ValueError():
    filepath = "test.1.fake"
    with raises(ValueError):
        _parse_and_validate_extension(Path(filepath), None, {".jpg": None})


def test_parse_mot_matching_extension_raises_ValueError():
    filepath = "test.jpg2"
    with raises(ValueError):
        _parse_and_validate_extension(Path(filepath), ".jpg", {".jpg": None})



from menpo.base import MenpoMissingDependencyError


try:
    from .vlfeat import dsift, fast_dsift, vector_128_dsift, hellinger_vector_128_dsift
except MenpoMissingDependencyError:
    pass

# Remove from scope
del MenpoMissingDependencyError



import numpy as np
from numpy.testing import assert_allclose
from pytest import raises

from menpo.image import MaskedImage, BooleanImage
from menpo.shape import PointCloud


def test_init_from_pointcloud_constrain_mask():
    pc = PointCloud(np.array([[5, 5], [5, 20], [20, 20]]))
    im = MaskedImage.init_from_pointcloud(pc, constrain_mask=True)
    assert im.shape == (15, 15)
    assert im.mask.n_true() == 120


def test_init_from_pointcloud_no_constrain_mask():
    pc = PointCloud(np.array([[5, 5], [5, 20], [20, 20]]))
    im = MaskedImage.init_from_pointcloud(pc, constrain_mask=False)
    assert im.shape == (15, 15)
    assert im.mask.n_true() == 225


def test_constrain_mask_to_landmarks_pwa():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks["box"] = PointCloud(
        np.array([[0.0, 0.0], [5.0, 0.0], [5.0, 5.0], [0.0, 5.0]])
    )
    new_img = img.constrain_mask_to_landmarks(group="box")

    example_mask = BooleanImage.init_blank((10, 10), fill=False)
    example_mask.pixels[0, :6, :6] = True
    assert new_img.mask.n_true() == 36
    assert_allclose(new_img.mask.pixels, example_mask.pixels)


def test_constrain_mask_to_landmarks_pwa_batched():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks["box"] = PointCloud(
        np.array([[0.0, 0.0], [5.0, 0.0], [5.0, 5.0], [0.0, 5.0]])
    )
    new_img = img.constrain_mask_to_landmarks(group="box", batch_size=2)

    example_mask = BooleanImage.init_blank((10, 10), fill=False)
    example_mask.pixels[0, :6, :6] = True
    assert new_img.mask.n_true() == 36
    assert_allclose(new_img.mask.pixels, example_mask.pixels)


def test_constrain_mask_to_landmarks_convex_hull():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks["box"] = PointCloud(
        np.array([[0.0, 0.0], [5.0, 0.0], [5.0, 5.0], [0.0, 5.0]])
    )
    new_img = img.constrain_mask_to_landmarks(
        group="box", point_in_pointcloud="convex_hull"
    )
    example_mask = BooleanImage.init_blank((10, 10), fill=False)
    example_mask.pixels[0, :6, 1:6] = True
    assert new_img.mask.n_true() == 30
    assert_allclose(new_img.mask.pixels, example_mask.pixels)


def test_constrain_mask_to_landmarks_callable():
    def bounding_box(_, indices):
        return np.ones(indices.shape[0], dtype=np.bool)

    img = MaskedImage.init_blank((10, 10))
    img.landmarks["box"] = PointCloud(
        np.array([[0.0, 0.0], [5.0, 0.0], [5.0, 5.0], [0.0, 5.0]])
    )
    new_img = img.constrain_mask_to_landmarks(
        group="box", point_in_pointcloud=bounding_box
    )
    example_mask = BooleanImage.init_blank((10, 10), fill=False)
    example_mask.pixels[0, :6, :6] = True
    assert new_img.mask.n_true() == 36
    assert_allclose(new_img.mask.pixels, example_mask.pixels)


def test_constrain_mask_to_landmarks_non_2d():
    img = MaskedImage.init_blank((10, 10, 10))
    img.landmarks["box"] = PointCloud(np.array([[0.0, 0.0, 0.0]]))
    with raises(ValueError):
        img.constrain_mask_to_landmarks()


def test_constrain_mask_to_landmarks_unknown_key():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks["box"] = PointCloud(np.array([[0.0, 0.0, 0.0]]))
    with raises(ValueError):
        img.constrain_mask_to_landmarks(point_in_pointcloud="unknown")


def test_erode():
    img = MaskedImage.init_blank((10, 10))
    img2 = img.erode()
    assert img2.mask.n_true() == 64
    img3 = img.erode(n_pixels=3)
    assert img3.mask.n_true() == 16


def test_constrain_mask_to_patches_around_landmarks_even():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks["box"] = PointCloud(
        np.array([[1.0, 1.0], [5.0, 1.0], [5.0, 5.0], [1.0, 5.0]])
    )
    new_img = img.constrain_mask_to_patches_around_landmarks((2, 2), group="box")
    assert new_img.mask.n_true() == 16
    assert_allclose(new_img.mask.pixels[:, 0, 0], True)
    assert_allclose(new_img.mask.pixels[:, 4:6, 0], True)
    assert_allclose(new_img.mask.pixels[:, 0, 4:6], True)
    assert_allclose(new_img.mask.pixels[:, 4:6, 4:6], True)


def test_constrain_mask_to_patches_around_landmarks_odd():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks["box"] = PointCloud(
        np.array([[2.0, 2.0], [5.0, 2.0], [5.0, 5.0], [2.0, 5.0]])
    )
    new_img = img.constrain_mask_to_patches_around_landmarks((3, 3), group="box")
    assert new_img.mask.n_true() == 36
    assert_allclose(new_img.mask.pixels[:, 1:3, 1:3], True)
    assert_allclose(new_img.mask.pixels[:, 4:7, 1:3], True)
    assert_allclose(new_img.mask.pixels[:, 1:3, 4:7], True)
    assert_allclose(new_img.mask.pixels[:, 4:7, 4:7], True)


def test_set_boundary_pixels():
    mask = np.ones((10, 10), dtype=np.bool)
    img = MaskedImage.init_blank((10, 10), mask=mask, fill=0.0, n_channels=1)
    new_img = img.set_boundary_pixels(value=2.0)
    assert new_img.mask.n_true() == 100
    assert ~np.allclose(img.pixels, new_img.pixels)
    assert_allclose(new_img.pixels[0, 1:-1, 1:-1], 0.0)
    assert_allclose(new_img.pixels[0, :, 0], 2.0)
    assert_allclose(new_img.pixels[0, 0, :], 2.0)
    assert_allclose(new_img.pixels[0, :, -1], 2.0)
    assert_allclose(new_img.pixels[0, -1, :], 2.0)


def test_dilate():
    img = MaskedImage.init_blank((10, 10))
    img = img.erode(n_pixels=3)
    img2 = img.dilate()
    assert img2.mask.n_true() == 32
    img3 = img.dilate(n_pixels=3)
    assert img3.mask.n_true() == 76


def test_init_from_rolled_channels():
    p = np.empty([50, 60, 3])
    im = MaskedImage.init_from_channels_at_back(p)
    assert im.n_channels == 3
    assert im.height == 50
    assert im.width == 60


def test_init_from_rolled_channels_masked():
    p = np.empty([50, 60, 3])
    example_mask = BooleanImage.init_blank((50, 60), fill=False)
    example_mask.pixels[0, :6, :6] = True

    im = MaskedImage.init_from_channels_at_back(p, mask=example_mask)
    assert im.n_channels == 3
    assert im.height == 50
    assert im.width == 60
    assert im.mask.n_true() == 36



import numpy as np

from .base import MultipleAlignment
from ..homogeneous import AlignmentSimilarity


class GeneralizedProcrustesAnalysis(MultipleAlignment):
    r"""
    Class for aligning multiple source shapes between them.

    After construction, the :map:`AlignmentSimilarity` transforms used to map
    each `source` optimally to the `target` can be found at `transforms`.

    Parameters
    ----------
    sources : `list` of :map:`PointCloud`
        List of pointclouds to be aligned.
    target : :map:`PointCloud`, optional
        The target :map:`PointCloud` to align each source to.
        If ``None``, then the mean of the sources is used.
    allow_mirror : `bool`, optional
        If ``True``, the Kabsch algorithm check is not performed, and mirroring
        of the Rotation matrix is permitted.

    Raises
    ------
    ValueError
        Need at least two sources to align
    """

    def __init__(self, sources, target=None, allow_mirror=False):
        super(GeneralizedProcrustesAnalysis, self).__init__(sources, target=target)
        initial_target = self.target
        self.transforms = [
            AlignmentSimilarity(source, self.target, allow_mirror=allow_mirror)
            for source in self.sources
        ]
        self.initial_target_scale = self.target.norm()
        self.n_iterations = 1
        self.max_iterations = 100
        self.converged = self._recursive_procrustes()
        if target is not None:
            self.target = initial_target

    def _recursive_procrustes(self):
        r"""
        Recursively calculates a procrustes alignment.
        """
        # Avoid circular imports
        from menpo.shape import mean_pointcloud, PointCloud
        from ..compositions import scale_about_centre

        if self.n_iterations > self.max_iterations:
            return False
        new_tgt = mean_pointcloud(
            [PointCloud(t.aligned_source().points, copy=False) for t in self.transforms]
        )
        # rescale the new_target to be the same size as the original about
        # it's centre
        rescale = scale_about_centre(
            new_tgt, self.initial_target_scale / new_tgt.norm()
        )
        rescale._apply_inplace(new_tgt)
        # check to see if we have converged yet
        delta_target = np.linalg.norm(self.target.points - new_tgt.points)
        if delta_target < 1e-6:
            return True
        else:
            self.n_iterations += 1
            for t in self.transforms:
                t.set_target(new_tgt)
            self.target = new_tgt
            return self._recursive_procrustes()

    def mean_aligned_shape(self):
        r"""
        Returns the mean of the aligned shapes.

        :type: :map:`PointCloud`
        """
        from menpo.shape import PointCloud

        return PointCloud(np.mean([t.target.points for t in self.transforms], axis=0))

    def mean_alignment_error(self):
        r"""
        Returns the average error of the recursive procrustes alignment.

        :type: `float`
        """
        return sum([t.alignment_error() for t in self.transforms]) / self.n_sources

    def __str__(self):
        if self.converged:
            return "Converged after %d iterations with av. error %f" % (
                self.n_iterations,
                self.mean_alignment_error(),
            )
        else:
            return "Failed to converge after %d iterations with av. error " "%f" % (
                self.n_iterations,
                self.mean_alignment_error(),
            )



import json
import itertools
import numpy as np


class _UTF8Encoder(json.JSONEncoder):
    def iterencode(self, obj, **kwargs):
        for chunk in json.JSONEncoder.iterencode(self, obj, **kwargs):
            yield chunk.encode("utf8")


def ljson_exporter(landmarks_object, file_handle, **kwargs):
    r"""
    Given a file handle to write in to (which should act like a Python `file`
    object), write out the landmark data. No value is returned.

    Writes out the LJSON format which is a verbose format that closely
    resembles the labelled point graph format. It describes semantic
    labels and connectivity between labels. The first axis of the format
    represents the image y-axis and is consistent with ordering within Menpo.

    Parameters
    ----------
    landmarks_object : dict or :map:`LandmarkManager`  or
        :map:`PointCloud` or subclass of :map:`PointCloud`
        The shape to write out.
    file_handle : `file`-like object
        The file to write in to
    """
    try:
        _ = landmarks_object.n_points
        landmark_dict = {"LJSON": landmarks_object}
    except AttributeError:
        # This should be a dict or a Landmark Manager.
        landmark_dict = landmarks_object

    # Add version string
    ljson = {"version": 3}
    groups = {}
    for key, pointcloud in landmark_dict.items():
        lg_json = pointcloud.tojson()

        # Convert nan values to None so that json correctly maps them to 'null'
        points = lg_json["landmarks"]["points"]
        # Flatten list
        try:
            ndim = len(points[0])
        except IndexError:
            ndim = 0
        filtered_points = [None if np.isnan(x) else x for x in itertools.chain(*points)]
        # Recreate tuples
        if ndim == 2:
            lg_json["landmarks"]["points"] = list(
                zip(filtered_points[::2], filtered_points[1::2])
            )
        elif ndim == 3:
            lg_json["landmarks"]["points"] = list(
                zip(filtered_points[::3], filtered_points[1::3], filtered_points[2::3])
            )
        else:
            lg_json["landmarks"]["points"] = []
        # append to the final ljson dict.
        groups[key] = lg_json

    ljson["groups"] = groups
    return json.dump(
        ljson,
        file_handle,
        indent=4,
        separators=(",", ": "),
        sort_keys=True,
        allow_nan=False,
        cls=_UTF8Encoder,
    )


def pts_exporter(pointcloud, file_handle, **kwargs):
    r"""
    Given a file handle to write in to (which should act like a Python `file`
    object), write out the landmark data. No value is returned.

    Writes out the PTS format which is a very simple format that does not
    contain any semantic labels. We assume that the PTS format has been created
    using Matlab and so use 1-based indexing and put the image x-axis as the
    first coordinate (which is the second axis within Menpo).

    Note that the PTS file format is only powerful enough to represent a
    basic pointcloud. Any further specialization is lost.

    Parameters
    ----------
    pointcloud : :map:`PointCloud` or subclass
        The pointcloud to write out.
    file_handle : `file`-like object
        The file to write in to
    """
    pts = pointcloud.points
    # Swap the x and y axis and add 1 to undo our processing
    # We are assuming (as on import) that the landmark file was created using
    # Matlab which is 1 based
    pts = pts[:, [1, 0]] + 1

    header = "version: 1\nn_points: {}\n{{".format(pts.shape[0])
    np.savetxt(
        file_handle,
        pts,
        delimiter=" ",
        header=header,
        footer="}",
        fmt="%.3f",
        comments="",
    )



from unittest.mock import Mock

from menpo.transform import Transform, TransformChain


def test_transformchain_apply():
    mocked_transform = Mock()
    mocked_transform._apply.return_value = 3
    mocked_transform2 = Mock()
    mocked_transform2._apply.return_value = 4
    transforms = [mocked_transform, mocked_transform2]
    tr = TransformChain(transforms)
    result = tr.apply(1)
    assert result == 4


def test_transformchain_composes_inplace_with():
    tr = TransformChain([])
    assert tr.composes_inplace_with == Transform


def test_transformchain_compose_before_composes_with():
    tr = TransformChain([])
    new_tr = tr.compose_before(Mock(spec=Transform))
    assert new_tr is not tr
    assert len(new_tr.transforms) is 1


def test_transformchain_compose_before_inplace_order():
    m1 = Mock(spec=Transform)
    m2 = Mock(spec=Transform)
    tr = TransformChain([m1])
    tr.compose_before_inplace(m2)
    assert tr.transforms[1] is m2


def test_transformchain_compose_after_inplace_order():
    m1 = Mock(spec=Transform)
    m2 = Mock(spec=Transform)
    tr = TransformChain([m1])
    tr.compose_after_inplace(m2)
    assert tr.transforms[0] is m2


def test_transformchain_compose_before_inplace_composes_with():
    tr = TransformChain([])
    ref = tr
    no_return = tr.compose_after_inplace(Mock(spec=Transform))
    assert no_return is None
    assert ref is tr
    assert len(tr.transforms) is 1


def test_transformchain_compose_after_composes_with():
    tr = TransformChain([])
    new_tr = tr.compose_after(Mock(spec=Transform))
    assert new_tr is not tr
    assert len(new_tr.transforms) is 1


def test_transformchain_compose_after_inplace_composes_with():
    tr = TransformChain([])
    ref = tr
    no_return = tr.compose_after_inplace(Mock(spec=Transform))
    assert no_return is None
    assert ref is tr
    assert len(tr.transforms) is 1



import numpy as np
from numpy.testing import assert_allclose

from menpo.image import Image
from menpo.image.rasterize import rasterize_landmarks_2d
from menpo.shape import PointCloud, PointUndirectedGraph

centre = PointCloud([[4.5, 4.5]])
line = PointUndirectedGraph(
    np.array([[2, 4.5], [8, 4.5]]), adjacency_matrix=np.array([[0, 1], [1, 0]])
)


def test_rasterize_matplotlib_basic():
    im = Image.init_blank([11, 11], fill=0, n_channels=1)
    im.landmarks["test"] = centre
    new_im = rasterize_landmarks_2d(
        im,
        group="test",
        render_lines=False,
        marker_style=".",
        marker_face_colour="r",
        marker_size=2,
        marker_edge_width=0,
        backend="matplotlib",
    )
    assert new_im.n_channels == 3
    assert new_im.shape == (11, 11)
    assert_allclose(new_im.pixels[:, 5, 5], [255, 0, 0])


def test_rasterize_pillow_basic():
    im = Image.init_blank([11, 11], fill=0, n_channels=3)
    im.landmarks["test"] = centre
    new_im = rasterize_landmarks_2d(
        im,
        group="test",
        render_lines=False,
        marker_style="s",
        marker_face_colour="r",
        marker_size=1,
        marker_edge_width=0,
        backend="pillow",
    )
    assert new_im.n_channels == 3
    assert new_im.shape == (11, 11)
    assert_allclose(new_im.pixels[0, 3:6, 3:6], 255)


def test_rasterize_pillow_basic_line():
    im = Image.init_blank([11, 11], fill=0, n_channels=3)
    im.landmarks["test"] = line
    new_im = rasterize_landmarks_2d(
        im,
        group="test",
        render_lines=True,
        line_width=1,
        line_colour="b",
        marker_style="s",
        marker_face_colour="r",
        marker_size=1,
        marker_edge_width=0,
        backend="pillow",
    )
    assert new_im.n_channels == 3
    assert_allclose(new_im.pixels[0, 1:4, 3:6], 255)
    assert_allclose(new_im.pixels[0, 7:-1, 3:6], 255)
    assert_allclose(new_im.pixels[2, 4:7, 4], 255)



import itertools
import json
import warnings
from collections import OrderedDict, namedtuple

import numpy as np
from scipy.sparse import csr_matrix

from menpo.shape import LabelledPointUndirectedGraph, PointCloud, PointUndirectedGraph
from menpo.transform import Scale

ASFPath = namedtuple(
    "ASFPath",
    [
        "path_num",
        "path_type",
        "xpos",
        "ypos",
        "point_num",
        "connects_from",
        "connects_to",
    ],
)


def asf_importer(filepath, asset=None, **kwargs):
    r"""
    Importer for the ASF file format.

    For images, the `x` and `y` axes are flipped such that the first axis is
    `y` (height in the image domain).

    Currently only open and closed path types are supported.

    Landmark labels:

    +---------+
    | label   |
    +=========+
    | all     |
    +---------+

    Parameters
    ----------
    filepath : `Path`
        Absolute filepath of the file.
    asset : `object`, optional
        An optional asset that may help with loading. This is unused for this
        implementation.
    \**kwargs : `dict`, optional
        Any other keyword arguments.

    Returns
    -------
    landmarks : `dict` {`str`: :map:`PointCloud`}
        Dictionary mapping landmark groups to menpo shapes

    References
    ----------
    .. [1] http://www2.imm.dtu.dk/~aam/datasets/datasets.html
    """
    with filepath.open("r") as f:
        landmarks = f.read()

    # Remove comments and blank lines
    landmarks = [l for l in landmarks.splitlines() if (l.rstrip() and not "#" in l)]

    # Pop the front of the list for the number of landmarks
    count = int(landmarks.pop(0))
    # Pop the last element of the list for the image_name
    image_name = landmarks.pop()

    xs = np.empty([count, 1])
    ys = np.empty([count, 1])
    connectivity = []

    # Only unpack the first 7 (the last 3 are always 0)
    split_landmarks = [ASFPath(*landmarks[i].split()[:7]) for i in range(count)]
    paths = [list(g) for k, g in itertools.groupby(split_landmarks, lambda x: x[0])]
    vert_index = 0
    for path in paths:
        if path:
            path_type = path[0].path_type
        for vertex in path:
            # Relative coordinates, will be scaled by the image size
            xs[vert_index, ...] = float(vertex.xpos)
            ys[vert_index, ...] = float(vertex.ypos)
            vert_index += 1
            # If True, isolated point
            if not (
                vertex.connects_from == vertex.connects_to
                and vertex.connects_to == vertex.point_num
            ):
                # Connectivity is defined by connects_from and connects_to
                # as well as the path_type:
                #   Bit 1: Outer edge point/Inside point
                #   Bit 2: Original annotated point/Artificial point
                #   Bit 3: Closed path point/Open path point
                #   Bit 4: Non-hole/Hole point
                # For now we only parse cases 0 and 4 (closed or open)
                connectivity.append((int(vertex.point_num), int(vertex.connects_to)))
        if path and path_type == "0":
            connectivity.append((int(path[-1].point_num), int(path[0].point_num)))

    connectivity = np.vstack(connectivity)
    points = np.hstack([ys, xs])
    if asset is not None:
        # we've been given an asset. As ASF files are normalized,
        # fix that here
        points = Scale(np.array(asset.shape)).apply(points)

    labels_to_masks = OrderedDict([("all", np.ones(points.shape[0], dtype=np.bool))])
    return {
        "ASF": LabelledPointUndirectedGraph.init_from_edges(
            points, connectivity, labels_to_masks
        )
    }


def pts_importer(filepath, image_origin=True, **kwargs):
    r"""
    Importer for the PTS file format. Assumes version 1 of the format.

    Implementations of this class should override the :meth:`_build_points`
    which determines the ordering of axes. For example, for images, the
    `x` and `y` axes are flipped such that the first axis is `y` (height
    in the image domain).

    Note that PTS has a very loose format definition. Here we make the
    assumption (as is common) that PTS landmarks are 1-based. That is,
    landmarks on a 480x480 image are in the range [1-480]. As Menpo is
    consistently 0-based, we *subtract 1* off each landmark value
    automatically.

    If you want to use PTS landmarks that are 0-based, you will have to
    manually add one back on to landmarks post importing.

    Landmark set label: PTS

    Parameters
    ----------
    filepath : `Path`
        Absolute filepath of the file.
    image_origin : `bool`, optional
        If ``True``, assume that the landmarks exist within an image and thus
        the origin is the image origin.
    \**kwargs : `dict`, optional
        Any other keyword arguments.

    Returns
    -------
    landmarks : `dict` {`str`: :map:`PointCloud`}
        Dictionary mapping landmark groups to menpo shapes
    """
    with filepath.open("r") as f:
        lines = [l.strip() for l in f.readlines()]

    line = lines[0]
    while not line.startswith("{"):
        line = lines.pop(0)

    xs = []
    ys = []
    for line in lines:
        if not line.strip().startswith("}"):
            xpos, ypos = line.split()[:2]
            xs.append(xpos)
            ys.append(ypos)

    xs = np.array(xs, dtype=np.float).reshape((-1, 1))
    ys = np.array(ys, dtype=np.float).reshape((-1, 1))

    # PTS landmarks are 1-based, need to convert to 0-based (subtract 1)
    if image_origin:
        points = np.hstack([ys - 1, xs - 1])
    else:
        points = np.hstack([xs - 1, ys - 1])

    return {"PTS": PointCloud(points, copy=False)}


def lm2_importer(filepath, **kwargs):
    r"""
    Importer for the LM2 file format from the bosphorus dataset. This is a 2D
    landmark type and so it is assumed it only applies to images.

    Landmark set label: LM2

    Landmark labels:

    +------------------------+
    | label                  |
    +========================+
    | outer_left_eyebrow     |
    | middle_left_eyebrow    |
    | inner_left_eyebrow     |
    | inner_right_eyebrow    |
    | middle_right_eyebrow   |
    | outer_right_eyebrow    |
    | outer_left_eye_corner  |
    | inner_left_eye_corner  |
    | inner_right_eye_corner |
    | outer_right_eye_corner |
    | nose_saddle_left       |
    | nose_saddle_right      |
    | left_nose_peak         |
    | nose_tip               |
    | right_nose_peak        |
    | left_mouth_corner      |
    | upper_lip_outer_middle |
    | right_mouth_corner     |
    | upper_lip_inner_middle |
    | lower_lip_inner_middle |
    | lower_lip_outer_middle |
    | chin_middle            |
    +------------------------+

    Parameters
    ----------
    filepath : `Path`
        Absolute filepath of the file.
    \**kwargs : `dict`, optional
        Any other keyword arguments.

    Returns
    -------
    landmarks : `dict` {`str`: :map:`PointCloud`}
        Dictionary mapping landmark groups to menpo shapes
    """
    with filepath.open("r") as f:
        landmarks = f.read()

    # Remove comments and blank lines
    landmark_text = [l for l in landmarks.splitlines() if (l.rstrip() and "#" not in l)]

    # First line says how many landmarks there are: 22 Landmarks
    # So pop it off the front
    num_points = int(landmark_text.pop(0).split()[0])
    labels = []

    # The next set of lines defines the labels
    labels_str = landmark_text.pop(0)
    if not labels_str == "Labels:":
        raise ValueError(
            "LM2 landmarks are incorrectly formatted. "
            "Expected a list of labels beginning with "
            "'Labels:' but found '{0}'".format(labels_str)
        )
    for i in range(num_points):
        # Lowercase, remove spaces and replace with underscores
        l = landmark_text.pop(0)
        l = "_".join(l.lower().split())
        labels.append(l)

    # The next set of lines defines the coordinates
    coords_str = landmark_text.pop(0)
    if not coords_str == "2D Image coordinates:":
        raise ValueError(
            "LM2 landmarks are incorrectly formatted. "
            "Expected a list of coordinates beginning with "
            "'2D Image coordinates:' "
            "but found '{0}'".format(coords_str)
        )
    xs = []
    ys = []
    for i in range(num_points):
        p = landmark_text.pop(0).split()
        xs.append(float(p[0]))
        ys.append(float(p[1]))

    xs = np.array(xs, dtype=np.float).reshape((-1, 1))
    ys = np.array(ys, dtype=np.float).reshape((-1, 1))

    # Flip the x and y
    points = np.hstack([ys, xs])
    # Create the mask whereby there is one landmark per label
    # (identity matrix)
    masks = np.eye(num_points).astype(np.bool)
    masks = np.vsplit(masks, num_points)
    masks = [np.squeeze(m) for m in masks]
    labels_to_masks = OrderedDict(zip(labels, masks))

    empty_adj_matrix = csr_matrix((num_points, num_points))
    return {
        "LM2": LabelledPointUndirectedGraph(points, empty_adj_matrix, labels_to_masks)
    }


def _ljson_parse_null_values(points_list):
    filtered_points = [
        np.nan if x is None else x for x in itertools.chain(*points_list)
    ]
    return np.array(filtered_points, dtype=np.float).reshape([-1, len(points_list[0])])


def _parse_ljson_v1(lms_dict):
    all_points = []
    labels = []  # label per group
    labels_slices = []  # slices into the full pointcloud per label
    offset = 0
    connectivity = []
    for group in lms_dict["groups"]:
        lms = group["landmarks"]
        labels.append(group["label"])
        labels_slices.append(slice(offset, len(lms) + offset))
        # Create the connectivity if it exists
        conn = group.get("connectivity", [])
        if conn:
            # Offset relative connectivity according to the current index
            conn = offset + np.asarray(conn)
            connectivity += conn.tolist()
        for p in lms:
            all_points.append(p["point"])
        offset += len(lms)

    # Don't create a PointUndirectedGraph with no connectivity
    points = _ljson_parse_null_values(all_points)
    n_points = points.shape[0]

    labels_to_masks = OrderedDict()
    # go through each label and build the appropriate boolean array
    for label, l_slice in zip(labels, labels_slices):
        mask = np.zeros(n_points, dtype=np.bool)
        mask[l_slice] = True
        labels_to_masks[label] = mask

    lmarks = LabelledPointUndirectedGraph.init_from_edges(
        points, connectivity, labels_to_masks
    )
    return {"LJSON": lmarks}


def _parse_ljson_v2(lms_dict):
    points = _ljson_parse_null_values(lms_dict["landmarks"]["points"])
    connectivity = lms_dict["landmarks"].get("connectivity")

    if connectivity is None and len(lms_dict["labels"]) == 0:
        lmarks = PointCloud(points)
    else:
        labels_to_mask = OrderedDict()  # masks into the pointcloud per label
        n_points = points.shape[0]
        for label in lms_dict["labels"]:
            mask = np.zeros(n_points, dtype=np.bool)
            mask[label["mask"]] = True
            labels_to_mask[label["label"]] = mask
        # Note that we can pass connectivity as None here and the edges will be
        # empty.
        lmarks = LabelledPointUndirectedGraph.init_from_edges(
            points, connectivity, labels_to_mask
        )

    return {"LJSON": lmarks}


def _parse_ljson_v3(lms_dict):
    all_lms = {}
    for key, lms_dict_group in lms_dict["groups"].items():
        points = _ljson_parse_null_values(lms_dict_group["landmarks"]["points"])
        connectivity = lms_dict_group["landmarks"].get("connectivity")
        labels_to_mask = OrderedDict()
        # TODO: create the metadata label!

        if len(lms_dict_group["labels"]) != 0:
            # masks into the pointcloud per label
            n_points = points.shape[0]
            for label in lms_dict_group["labels"]:
                mask = np.zeros(n_points, dtype=np.bool)
                mask[label["mask"]] = True
                labels_to_mask[label["label"]] = mask

        # Note that we can pass connectivity as None here and the edges will be empty.
        graph_cls = (
            LabelledPointUndirectedGraph if labels_to_mask else PointUndirectedGraph
        )
        all_lms[key] = graph_cls.init_from_edges(points, connectivity, labels_to_mask)

    return all_lms


_ljson_parser_for_version = {1: _parse_ljson_v1, 2: _parse_ljson_v2, 3: _parse_ljson_v3}


def ljson_importer(filepath, **kwargs):
    r"""
    Importer for the Menpo JSON format. This is an n-dimensional
    landmark type for both images and meshes that encodes semantic labels in
    the format.

    Landmark set label (v1, v2): JSON
    Landmark set label (v3): As defined in the file

    Parameters
    ----------
    filepath : `Path`
        Absolute filepath of the file.
    \**kwargs : `dict`, optional
        Any other keyword arguments.

    Returns
    -------
    landmarks : `dict` {`str`: :map:`PointCloud`}
        Dictionary mapping landmark groups to menpo shapes
    """
    with filepath.open("r") as f:
        lms_dict = json.load(f, object_pairs_hook=OrderedDict)
    version = lms_dict.get("version")
    parser = _ljson_parser_for_version.get(version)

    if parser is None:
        raise ValueError(
            "{} has unknown version {} - must be "
            "1, or 2 or 3.".format(filepath, version)
        )
    if version != 3:
        from menpo.base import MenpoDeprecationWarning

        warnings.warn(
            "LJSON v{} is deprecated. export_landmark_file() will "
            "only save out LJSON v3 files. Please convert all LJSON "
            "files to v3 by importing into Menpo and re-exporting to "
            "overwrite the files.".format(version),
            MenpoDeprecationWarning,
        )

    return parser(lms_dict)
